#include "split_compounds.h"

#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>
#include <vector>
#include <string>

/* Location of data files for Stava*/
// #define LIBPATH "/home/jsh/servergranskafiles/stava/GPLstava_extra/lib/"
#define LIBPATH STAVALEXICONFOLDER

#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <strings.h>

/* Assuming the text is Swedish text, this should return 1 if the
   encoding is UTF-8, and it should return 0 if we find something that
   looks like latin-1 */
int looksLikeUTF(const char *s) {
  if(s == NULL) {
    return 1;
  }

  int last = 1;
  for(int i = 0; s[i] != '\0'; i++) {
    if(s[i] < 0 && last > 0) {
      if(s[i] == -27
	 || s[i] == -28
	 || s[i] == -10
	 || s[i] == -59
	 || s[i] == -60
	 || s[i] == -42) {
	return 0;
      }
    }
    if(s[i] < 0 && last == -61) {
      return 1;
    }
    last = s[i];
  }
  
  return 1;
}

/* Converts UTF-8 to Latin-1 correctly if we assume the string
   contents are Swedish text. Based on code example from Mark Ransom */
void utf2latin1(std::string &s) {
  
  int n = s.size();
  std::string temp = "";
  temp.reserve(n);

  unsigned int codepoint = 0;

  int j = 0;
  for (int i = 0; s[i] != 0; ) {
    /* unsigned char c = static_cast<unsigned char>(s[i]); */
    unsigned char c = (unsigned char)(s[i]);
    if (c <= 0x7f)
      codepoint = c;
    else if (c <= 0xbf)
      codepoint = (codepoint << 6) | (c & 0x3f);
    else if (c <= 0xdf)
      codepoint = c & 0x1f;
    else if (c <= 0xef)
      codepoint = c & 0x0f;
    else
      codepoint = c & 0x07;

    i++;
    
    if (((s[i] & 0xc0) != 0x80) && (codepoint <= 0x10ffff)) {
      if (codepoint <= 255) {
	/* tempBuf[j++] = static_cast<char>(codepoint); */
	temp += (char)(codepoint);
      } else {
	// do whatever you want for out-of-bounds characters
	temp += 'X'; // Store something away so we can convert back later? TODO: Jonas
      }
    }
  }

  s = temp;
}

int REQUESTS = 0;
int WORDS = 0;
int COMPOUNDS = 0;

const int SOCKETBUFSIZE = 1000000;
void read_data(int socket, std::vector<std::string> &words) {
  char buf[SOCKETBUFSIZE+1];
  buf[0] = 0;
  bool done = 0;
  std::string text = "";

  while(!done) {
    int len = 0, offs = 0;
    while(offs < SOCKETBUFSIZE && !done) {
      len = read(socket, buf+offs, SOCKETBUFSIZE - offs);
      if(len > 0) {
	buf[SOCKETBUFSIZE] = 0;
	char *end = strstr(buf, "ENDQ");
	if(end) {
	  *end = 0;
	  done = 1;
	} else {
	  offs += len;
	}
      } else {
	buf[offs] = 0;
	done = 1;
      }

      if(offs >= SOCKETBUFSIZE) {
	buf[SOCKETBUFSIZE-1] = 0;
	done = 1;
      }
    }
    text += buf;
    int tl = text.length() - 10;
    if(tl < 0)
      tl = 0;
    if(text.substr(tl).find("ENDQ") != std::string::npos) {
      done = 1;
    }
  }

  if(looksLikeUTF(text.c_str())) {
    utf2latin1(text);
  }

  std::istringstream is(text);
  while(is) {
    std::string word;
    is >> word;
    if(word == "ENDQ") {
      break;
    }
    if(word.length() > 0) {
      words.push_back(word);
    }
  }
}

// taken from a socket primer
int establish(int portnum)
{ 
  const int MAXHOSTNAME = 1000;
  char myname[MAXHOSTNAME+1];
  int s;
  struct sockaddr_in sa;
  struct hostent *hp;
  
  bzero(&sa,sizeof(struct sockaddr_in));
  gethostname(myname,MAXHOSTNAME);
  hp= gethostbyname(myname);
  if (hp == NULL)
    return -1;
  sa.sin_family= hp->h_addrtype; /* this is our host address */
  sa.sin_port= htons(portnum); /* this is our port number */
  if ((s = socket(AF_INET,SOCK_STREAM,0)) < 0)
    return -1;
  if (bind(s,(struct sockaddr *)&sa,sizeof(sa)) < 0) {
    close(s);
    return -1;
  }
  listen(s, 5); /* max # of queued connects */
  return s;
}

// taken from a socket primer
int get_connection(int s // socket
		   )
{ 
  struct sockaddr_in isa;
  socklen_t i;
  int t;
  
  i = sizeof(isa);
  getsockname(s,(struct sockaddr *)&isa,&i);
  
  if ((t = accept(s,(struct sockaddr *)&isa,&i)) < 0)   /* accept connection if there is one */
    return -1;
  return t;
}

int main(int argc, char **argv) {
  size_t temp;
  
  if(argc != 7) {
    std::cerr << argv[0] << " <head freqs> <tail freqs> <ngram-stats> <port no> <tagger lexicon> <tag tag stats>\n";
    return 0;
  }

  /* handler for child processes */
  signal(SIGCHLD, SIG_IGN);

  int portno = atoi(argv[4]);
  int server_socket;
  if ((server_socket = establish(portno)) < 0) {  /* plug in the phone */
    std::cerr << "could not open socket for listening\n";
    exit(1);
  }

  init(LIBPATH, argv[1], argv[2], argv[3], argv[5], argv[6]);

  std::vector<std::string> words, result;
  
  while(1) {
    std::cerr << "waiting...\n";
    int client_socket = 0;
    if ((client_socket = get_connection(server_socket)) < 0) {
      if (errno == EINTR)
	continue;
      std::cerr << "error on accept\n";
      exit(1);
    }
    
    // fork all client calls, that way, if they hang or crash,
    // the server keeps working
    std::ostringstream buf;
    switch(fork()) {
    case -1 : // error
      std::cerr << "error on fork()\n";
      exit(1);
    case 0 : // child
      close(server_socket);
      words.clear();
      alarm(60*2); // if we still haven't finished in 2 minutes, die...

      read_data(client_socket, words);
      alarm(0); // remove alarm
      
      result.clear();
      split_compounds(words, result);
      
      buf << "<reply>\n";
      for(int i = 0; i < words.size(); i++) {
	buf << words[i];
	if(result[i] != "") {
	  buf << "\t" << result[i];
	  COMPOUNDS++;
	}
	WORDS++;
	buf << "\n";
      }
      buf << "</reply>\n";

      temp = write(client_socket, buf.str().c_str(), buf.str().length() * sizeof(char));
      
      close(client_socket);
      exit(0); // child is finished
    default : // parent
      REQUESTS++;
      close(client_socket);
      wait(NULL); // We do not want to spam out too many processes at once, so wait for child to finish
    }
  }
  return 0;
}
